﻿using Nemerle.Collections;
using Nemerle.Imperative;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;
using SCG = System.Collections.Generic;


namespace neorbe.dna
{
  public variant DnaBlock
  {
    | Blob { data : array[byte] }
    | Ref { blob : Blob; first : int; last : int; }
    
    public Size : int
    {
      get
      {
        match (this)
        {
          | Blob(data) => data.Length
          | Ref(_, first, last) => last - first
        }
      }
    }
  }

  public class Dna : SCG.IEnumerable[byte]
  {
    mutable blocks : list[DnaBlock];
    mutable period = 0;
    
    public this(init : Vector[byte])
    {
      blocks = [DnaBlock.Blob(init.ToArray())];
    }
  
    public Get(index : list[DnaBlock]) : byte
    {
      match (index)
      {
        | DnaBlock.Blob(data)::_ => data[0]
        | DnaBlock.Ref(blob, first, _)::_ => blob.data[first]
        | [] => throw Exception()
      }
    }
    
    public Start : list[DnaBlock]
    {
      get
      {
        blocks
      }
      set
      {
        blocks = value;
      }
    }
    
    public HasItem(index : list[DnaBlock]) : bool
    {
      match (index)
      {
        | _::_ => true
        | []   => false
      }
    }
    
    public Size : int
    {
      get
      {
        mutable sz = 0;
        foreach (block in blocks)
        {
          sz += match (block)
          {
            | Blob(data) => data.Length
            | Ref(_, from, to) => to - from
          }          
        }
        sz
      }
    }
    
    public MakeOffset(offset : int, cursor : list[DnaBlock]) : list[DnaBlock]
    {
      make_offset(offset, cursor)
    }
    
    make_offset(offset : int, cursor : list[DnaBlock]) : list[DnaBlock]
    {
      if (offset <= 0)
        cursor
      else
        match (cursor)
        {
          | head::tail =>
            def sz = head.Size;
            if (sz <= offset)
              make_offset(offset - sz, tail)
            else
              match (head)
              {
                | Blob(_) as blob => DnaBlock.Ref(blob, offset, blob.data.Length) :: tail
                | Ref(blob, first, last) => DnaBlock.Ref(blob, first + offset, last) :: tail
              }
          | [] => []
        }
    }
    
    public RemoveFromStart(count : int) : void
    {
      blocks = make_offset(count, blocks);
    }
    
    public GetEnumerator() : SCG.IEnumerator[byte]
    {
      foreach (block in blocks)
      {
        match (block)
        {
          | Blob(data) =>
            foreach (c in data)
              yield c;
          | Ref(blob, from, to) =>
            for (mutable j = from; j < to; ++j)
              yield blob.data[j];
        }
      }
    }
    
    GetEnumeratorOld() : System.Collections.IEnumerator
      implements System.Collections.IEnumerable.GetEnumerator
    {
      GetEnumerator()
    }
    
    public Take(count : int) : SCG.IEnumerable[byte]
    {
      mutable i = 0;
      foreach (b in this)
      {
        when (i >= count) break;
        ++i;
        yield b;
      }
    }
        
    public TakeList(count : int) : list[char]
    {
      def sb = StringBuilder();
      foreach (b in Take(count))
        _ = sb.Append(b :> char);
      sb.ToString().ToList();
    }
    
    public MatchString(text : string) : bool
    {
      def toMatch = Vector(Take(text.Length));
      if (toMatch.Count != text.Length)
        false
      else
      {
        for (mutable i = 0; i < text.Length; ++i)
          when (text[i] :> byte != toMatch[i])
            return false;
        true
      }
    }
    
    public GetBlocksFromCursor(cursor : list[DnaBlock], count : int) : list[DnaBlock]
    {
      def get_blocks_core(cursor, count, acc = [])
      {
        if (count <= 0)
          acc.Reverse()
        else  
          match (cursor)
          {
            | [] => acc.Reverse()
            | head::tail =>
              def sz = head.Size;
              if (sz > count)
              {
                def r = match (head)
                {
                  | DnaBlock.Blob(_) as blob => DnaBlock.Ref(blob, 0, count)
                  | DnaBlock.Ref(blob, from, _) => DnaBlock.Ref(blob, from, from + count)
                }
                (r :: acc).Reverse()
              }
              else
                get_blocks_core(tail, count - sz, head :: acc)
          }
      }
      get_blocks_core(cursor, count)
    }
    
    public AppendToStart(newData : list[DnaBlock]) : void
    {
      blocks = newData.Append(blocks);
      ++period;
      when (period > 6)
      {
        /*when (blocks.Length > 128)
          blocks = [DnaBlock.Blob(Vector(this).ToArray())];*/
        defragment();
        period = 0;
      }
    }
    
    defragment() : void
    {
      def result = Vector();
      def accum = Vector();
      def flush()
      {
        when (accum.Count > 0)
        {
          result.Add(DnaBlock.Blob(accum.ToArray()));
          accum.Clear();
        }
      }
      foreach (block in blocks)
      {
        if (block.Size > 16384)
        {
          flush();
          result.Add(block);
        }
        else
        {
          accum.AddRange(enum_block(block));
          /*when (accum.Count > 262144)
            flush();*/
        }
      }
      flush();
      blocks = result.ToList();
    }
    
    static enum_block(block : DnaBlock) : SCG.IEnumerable[byte]
    {
      match (block)
      {
        | Blob(data) => data : SCG.IEnumerable[byte]
        | Ref(_, _, _) as r =>
          enum_ref(r)
      }
    }
    
    static enum_ref(r : DnaBlock.Ref) : SCG.IEnumerable[byte]
    {
      for (mutable i = r.first; i < r.last; ++i)
        yield r.blob.data[i];
    }
    
    public override ToString() : string
    {
      def sb = StringBuilder();
      foreach (c in this.Take(50))
        _ = sb.Append(c :> char);
      sb.ToString()
    }
  }
  
  public variant PatternSymbol
  {
    | Item { val : byte; }
    | Skip { n : int; }
    | Sequence { seq : Vector[byte]; }
    | Open
    | Close
  }
  
  public type Pattern = Vector[PatternSymbol];
  
  public variant TemplateSymbol
  {
    | Item { val : byte; }
    | Protected { reference : int; protection : int; }
    | Length { n : int; }
  }
  
  public type Template = Vector[TemplateSymbol];
  
  public class DnaException : Exception
  {
  }
  
  public class DnaProcessor
  {
    dna : Dna;
    writer : TextWriter;
    
    public this(dna : Dna, writer : TextWriter)
    {
      this.dna = dna;
      this.writer = writer;
    }
    
    public CurrentDna : Dna
    {
      get
      {
        dna
      }
    }
    
    public Process() : void
    {
      mutable iterations = 0;
      def sw = Stopwatch();
      sw.Start();
      while (true)
      {
        try
        {
          def p = DecodePattern();
          def t = DecodeTemplate();
          MatchReplace(p, t);
        }
        catch
        {
          _ is DnaException => break;
        }
        if (iterations < 10000)
          ++iterations
        else
        {
          sw.Stop();
          Console.Error.WriteLine(sw.ElapsedMilliseconds);
          iterations = 0;
          sw.Reset();
          sw.Start();
        }
      }
    }
    
    public DecodePattern() : Pattern
    {
      def pattern = Vector();
      mutable nesting = 1;
      
      while (true)
      {
        def l = dna.TakeList(3);
        match (l)
        {
          | 'C' :: _ =>
            dna.RemoveFromStart(1);
            pattern.Add(PatternSymbol.Item('I' :> byte));
          | 'F' :: _ =>
            dna.RemoveFromStart(1);
            pattern.Add(PatternSymbol.Item('C' :> byte));
          | 'P' :: _ =>
            dna.RemoveFromStart(1);
            pattern.Add(PatternSymbol.Item('F' :> byte));   
          | 'I' :: 'C' :: _ =>
            dna.RemoveFromStart(2);
            pattern.Add(PatternSymbol.Item('P' :> byte));
          | 'I' :: 'P' :: _ =>
            dna.RemoveFromStart(2);
            def n = DecodeNatural();
            pattern.Add(PatternSymbol.Skip(n));
          | 'I' :: 'F' :: _ =>
            dna.RemoveFromStart(3);
            def c = Vector();
            c.AddRange(DecodeConst());
            pattern.Add(PatternSymbol.Sequence(c));
          | 'I' :: 'I' :: 'P' :: _ =>
            dna.RemoveFromStart(3);
            nesting++;
            pattern.Add(PatternSymbol.Open());  
          | 'I' :: 'I' :: 'C' :: _
          | 'I' :: 'I' :: 'F' :: _ =>
            dna.RemoveFromStart(3);
            if (nesting == 1)
              break;
            else
            {
              nesting--;
              pattern.Add(PatternSymbol.Close());
            }
          | 'I' :: 'I' :: 'I' :: _ =>
            dna.RemoveFromStart(3);
            for (mutable i = 0; i < 7; ++i)
            {
              when (!dna.HasItem(dna.Start))
              {
                Console.Error.WriteLine($"DNA length: $(dna.ToString().Length)");
                break;
              }
              writer.Write(dna.Get(dna.Start) :> char);
              writer.Flush();
              dna.RemoveFromStart(1);
            }
          | _ =>
            throw DnaException()
        }
      }
      pattern
    }
    
    public DecodeNatural() : int
    {
      mutable n = 0;
      mutable bit = 1l;
      while (true)
      {
        when (!dna.HasItem(dna.Start))
          throw DnaException();
        def c = dna.Get(dna.Start);
        dna.RemoveFromStart(1);
        when (bit < 0x3FFFFFFF)
        {
          match (c :> char)
          {
            | 'P' => break;
            | 'I' | 'F' => ()
            | 'C' => n += (bit :> int)
            | _ => throw Exception()
          }
          bit *= 2l;
        }
      }
      if (bit >= 0x3FFFFFFF)
        0x3FFFFFFF
      else
        n
    }
    
    public DecodeConst() : Vector[byte]
    {
      def result = Vector();
      while (true)
      {
        def l = dna.TakeList(2);
        match (l)
        {
          | 'C' :: _ =>
            dna.RemoveFromStart(1);
            result.Add('I' :> byte);
          | 'F' :: _ =>
            dna.RemoveFromStart(1);
            result.Add('C' :> byte);
          | 'P' :: _ =>
            dna.RemoveFromStart(1);
            result.Add('F' :> byte);
          | 'I' :: 'C' :: _ =>
            dna.RemoveFromStart(2);
            result.Add('P' :> byte);
          | _ =>
            break;
        }
      }
      result
    }
    
    public DecodeTemplate() : Template
    {
      def t = Vector();
      while (true)
      {
        def l = dna.TakeList(3);
        match (l)
        {
          | 'C' :: _ =>
            dna.RemoveFromStart(1);
            t.Add(TemplateSymbol.Item('I' :> byte));  
          | 'F' :: _ =>
            dna.RemoveFromStart(1);
            t.Add(TemplateSymbol.Item('C' :> byte));  
          | 'P' :: _ =>
            dna.RemoveFromStart(1);
            t.Add(TemplateSymbol.Item('F' :> byte));  
          | 'I' :: 'C' :: _ =>
            dna.RemoveFromStart(2);
            t.Add(TemplateSymbol.Item('P' :> byte));  
          | 'I' :: 'P' :: _
          | 'I' :: 'F' :: _ =>
            dna.RemoveFromStart(2);
            def l = DecodeNatural();
            def n = DecodeNatural();
            t.Add(TemplateSymbol.Protected(n, l));  
          | 'I' :: 'I' :: 'C' :: _
          | 'I' :: 'I' :: 'F' :: _ =>
            dna.RemoveFromStart(3);
            break;
          | 'I' :: 'I' :: 'P' :: _ =>
            dna.RemoveFromStart(3);
            def n = DecodeNatural();
            t.Add(TemplateSymbol.Length(n));
          | 'I' :: 'I' :: 'I' :: _ =>
            dna.RemoveFromStart(3);
            for (mutable i = 0; i < 7; ++i)
            {
              when (!dna.HasItem(dna.Start))
                break;
              writer.Write(dna.Get(dna.Start) :> char);
              writer.Flush();
              dna.RemoveFromStart(1);
            }
          | _ =>
            throw DnaException()
        }          
      }
      t
    }
    
    public MatchReplace(pat : Pattern, templ : Template) : void
    {
      mutable i = dna.Start;
      mutable index = 0;
      def stack = SCG.Stack();
      def env = Vector();
      foreach (p in pat)
      {
        match (p)
        {
          | Item(val) =>
            if (dna.HasItem(i) && dna.Get(i) == val)
            {
              i = dna.MakeOffset(1, i);
              ++index;
            }
            else
              return;
          | Skip(n) =>
            when (n != 0)
            {
              i = dna.MakeOffset(n - 1, i);
              when (!dna.HasItem(i))            
                return;
              i = dna.MakeOffset(1, i);
              index += n;
            }
          | Sequence(val) =>
            match (match_seq(i, val))
            {
              | option.None() => return;
              | option.Some((n, sz)) => i = n; index += sz;
            }
          | Open => stack.Push(index)
          | Close =>
            def last = stack.Pop();
            def sz = index - last;
            def s = dna.MakeOffset(last, dna.Start);
            def r = dna.GetBlocksFromCursor(s, sz);
            env.Add(r);
        }
      }
      dna.Start = i;
      Replace(templ, env)
    }
    
    public Replace(templ : Template, env : Vector[list[DnaBlock]]) : void
    {
      def result = Vector();
      def bytes = Vector();
      def flush_bytes()
      {
        when (bytes.Count > 0)
          result.Add(DnaBlock.Blob(bytes.ToArray()));
        bytes.Clear();
      }
      foreach (t in templ)
      {
        match (t)
        {
          | Item(val) =>
            bytes.Add(val);
          | Protected(n, 0) =>
            when (n >= 0 && n < env.Count)
            {
              flush_bytes();
              result.AddRange(env[n]);
            }
          | Protected(n, l) =>
            when (n >= 0 && n < env.Count)
              bytes.AddRange(protect(l, get_content(env[n])));
          | Length(n) =>
            def sz = if (n >= 0 && n < env.Count) get_size(env[n]) else 0;
            bytes.AddRange(to_number(sz));
        }
      }
      flush_bytes();
      dna.AppendToStart(result.ToList())
    }
    
    static get_size(blocks : list[DnaBlock]) : int
    {
      blocks.Map(x => x.Size).FoldLeft(0, _ + _)
    }
    
    static get_content(src : SCG.IEnumerable[DnaBlock]) : SCG.IEnumerable[byte]
    {
      foreach (block in src)
      {
        match (block)
        {
          | DnaBlock.Blob(data) =>
            foreach (b in data)
              yield b;
          | DnaBlock.Ref(blob, from, to) =>
            for (mutable i = from; i < to; ++i)
              yield blob.data[i];
        }
      }
    }
    
    static protect(l : int, d : SCG.IEnumerable[byte]) : Vector[byte]
    {
      def quote(d)
      {
        def result = Vector();
        foreach (c in d)
        {
          match (c :> char)
          {
            | 'I' => result.Add('C' :> byte)
            | 'C' => result.Add('F' :> byte)
            | 'F' => result.Add('P' :> byte)
            | 'P' => result.Add('I' :> byte); result.Add('C' :> byte);
            | _ => throw Exception()
          }
        }
        result
      }
      mutable r = Vector(d);
      for (mutable i = 0; i < l; ++i)
        r = quote(r);
      r
    }
    
    static to_number(n : int) : Vector[byte]
    {
      def result = Vector();      
      mutable k = n;
      while (k > 0)
      {
        if (k % 2 == 0)
          result.Add('I' :> byte);
        else
          result.Add('C' :> byte);
        k /= 2;
      }
      result.Add('P' :> byte);
      result
    }
    
    match_seq(from : list[DnaBlock], pat : Vector[byte]) : option[list[DnaBlock] * int]
    {
      def automata = Vector();
      mutable i = from;
      mutable size = 0;
      while (dna.HasItem(i))
      {
        def c = dna.Get(i);
        automata.Add(0);
        mutable j = 0;
        while (j < automata.Count)
        {
          def patIndex = automata[j];
          if (patIndex < pat.Count && c == pat[patIndex])
          {
            automata[j] = patIndex + 1;
            ++j;
          }
          else
          {
            if (patIndex >= pat.Count)
              return option.Some((i, size));
            else
              automata.RemoveAt(j);   
          }
        }
        i = dna.MakeOffset(1, i);
        ++size;
      }
      foreach (index in automata)
        when (index >= pat.Count) return option.Some([], size);
      option.None()
    }
  }
}